@@ -22,6 +22,24 @@
 
 type fd = Unix.file_descr
 
+external poll : Unix.file_descr array -> Unix.file_descr array -> Unix.file_descr array -> float -> (Unix.file_descr array * Unix.file_descr array * Unix.file_descr array) = "caml_poll"
+
+let poll r w e timeout =
+  let r = Array.of_list r in
+  let w = Array.of_list w in
+  let e = Array.of_list e in
+  let (r, w, e) =
+    poll r w e timeout
+  in
+  (Array.to_list r,
+   Array.to_list w,
+   Array.to_list e)
+
+let select, select_fname =
+  match Sys.os_type with
+    | "Unix" -> poll, "poll"
+    | _ -> Unix.select, "select"
+
 (** [remove f l] is like [List.find f l] but also returns the result of removing
   * the found element from the original list. *)
 let remove f l =
@@ -61,6 +79,7 @@ type 'a scheduler =
   mutable queues : Condition.t list;
   queues_m : Mutex.t;
   mutable stop : bool;
+  stop_m : Mutex.t
 }
 
 let clear_tasks s = 
@@ -82,6 +101,7 @@ let create ?(compare=compare) () =
     queues = [];
     queues_m = Mutex.create ();
     stop = false;
+    stop_m = Mutex.create ()
   }
 
 let wake_up s = ignore (Unix.write s.in_pipe (Bytes.of_string "x") 0 1)
@@ -152,8 +172,14 @@ end
 open Task
 
 let stop s = 
-  s.stop <- true ;
-  wake_up s
+  clear_tasks s;
+  Mutex.lock s.stop_m;
+  s.stop <- true;
+  Mutex.unlock s.stop_m;
+  wake_up s;
+  Mutex.lock s.ready_m;
+  List.iter Condition.signal s.queues;
+  Mutex.unlock s.ready_m
 
 let tmp = Bytes.create 1024
 
@@ -175,11 +201,11 @@ let process s log =
         let timeout = 
           if e.t = infinity then -1. else max 0. (e.t -. (time ())) 
         in
-        log (Printf.sprintf "Enter select at %f, timeout %f (%d/%d/%d)."
-             (time ()) timeout
+        log (Printf.sprintf "Enter %s at %f, timeout %f (%d/%d/%d)."
+             select_fname (time ()) timeout
            (List.length e.r) (List.length e.w) (List.length e.x)) ;
-        let r,w,x = Unix.select e.r e.w e.x timeout in
-        log (Printf.sprintf "Left select at %f (%d/%d/%d)." (time ())
+        let r,w,x = select e.r e.w e.x timeout in
+        log (Printf.sprintf "Left %s at %f (%d/%d/%d)." select_fname (time ())
                (List.length r) (List.length w) (List.length x)) ;
         r,w,x
       with
@@ -254,6 +280,9 @@ let exec s (priorities:'a->bool) =
       true
   with Not_found -> false
 
+exception Queue_stopped
+exception Queue_processed
+
  (** Main loop for queues. *)
 let queue ?log ?(priorities=fun _ -> true) s name = 
   let log = 
@@ -269,49 +298,63 @@ let queue ?log ?(priorities=fun _ -> true) s name =
       log (Printf.sprintf "Queue #%d starting..." (List.length s.queues)) ;
       c
   in
-    (* Try to process ready tasks, otherwise try to become the master,
-     * or be a slave and wait for the master to get some more ready tasks. *)
-    while not s.stop do
-      (* Lock the ready tasks until the queue has a task to proceed,
-       * *or* is really ready to restart on its condition, see the 
-       * Condition.wait call below for the atomic unlock and wait. *)
-      Mutex.lock s.ready_m ;
-      log (Printf.sprintf "There are %d ready tasks." (List.length s.ready)) ;
-      if exec s priorities then () else 
-          let wake () = 
-            (* Wake up other queues 
-	     * if there are remaining tasks *)
-            if s.ready <> [] then begin
-	        Mutex.lock s.queues_m ;
-                List.iter (fun x -> if x <> c then Condition.signal x) 
-		          s.queues ;
-	        Mutex.unlock s.queues_m 
-	      end ;
-          in
-          if Mutex.try_lock s.select_m then begin
-	    (* Processing finished for me
-	     * I can unlock ready_m now.. *)
+  (* Try to process ready tasks, otherwise try to become the master,
+   * or be a slave and wait for the master to get some more ready tasks. *)
+  let run () =
+    Mutex.lock s.stop_m;
+    let stop = s.stop in
+    Mutex.unlock s.stop_m;
+    if stop then raise Queue_stopped;
+    (* Lock the ready tasks until the queue has a task to proceed,
+     * *or* is really ready to restart on its condition, see the 
+     * Condition.wait call below for the atomic unlock and wait. *)
+    Mutex.lock s.ready_m ;
+    log (Printf.sprintf "There are %d ready tasks." (List.length s.ready)) ;
+    if exec s priorities then raise Queue_processed;
+    let wake () = 
+      (* Wake up other queues if there are remaining tasks *)
+      if s.ready <> [] then
+       begin
+	      Mutex.lock s.queues_m ;
+        List.iter (fun x -> if x <> c then Condition.signal x) 
+		    s.queues ;
+	      Mutex.unlock s.queues_m 
+	     end ;
+    in
+    if Mutex.try_lock s.select_m then begin
+	   (* Processing finished for me
+	    * I can unlock ready_m now.. *)
 	    Mutex.unlock s.ready_m ;
-            process s log ;
-            Mutex.unlock s.select_m ;
+      process s log ;
+      Mutex.unlock s.select_m ;
 	    Mutex.lock s.ready_m ;
-            wake () ;
+      wake () ;
 	    Mutex.unlock s.ready_m
-          end else begin
-              (* We use s.ready_m mutex here.
-	       * Hence, we avoid race conditions
-	       * with any other queue being processing
-	       * a task that would create a new task: 
-	       * without this mutex, the new task may not be 
-	       * notified to this queue if it is going to sleep
-	       * in concurrency..
-	       * It also avoid race conditions when restarting 
-	       * queues since s.ready_m is locked until all 
-	       * queues have been signaled. *)
-              Condition.wait c s.ready_m;
-              Mutex.unlock s.ready_m
-            end
-    done
+    end else begin
+    (* We use s.ready_m mutex here.
+  	 * Hence, we avoid race conditions
+  	 * with any other queue being processing
+  	 * a task that would create a new task: 
+  	 * without this mutex, the new task may not be 
+	   * notified to this queue if it is going to sleep
+	   * in concurrency..
+	   * It also avoid race conditions when restarting 
+	   * queues since s.ready_m is locked until all 
+	   * queues have been signaled. *)
+     Condition.wait c s.ready_m;
+     Mutex.unlock s.ready_m
+    end
+   in
+   let rec f () =
+    begin
+     try run () with
+      | Queue_processed -> ()
+    end;
+    (f [@tailcall]) ()
+   in
+   try
+     f ()
+   with Queue_stopped -> ()
 
 module Async =
 struct
@@ -588,6 +631,11 @@ struct
             0,fun _ _ -> 0
     in
     let unix_socket = Transport.sock (socket:Transport.t) in
+    let exec () =
+      if Sys.os_type = "Win32" then
+        Unix.clear_nonblock unix_socket;
+      exec ()
+    in
     let events,check_timeout =
       match timeout with
         | None -> [`Write unix_socket], fun _ -> false
@@ -611,18 +659,30 @@ struct
           (exec () ; [])
       end
      with
+       | Unix.Unix_error(Unix.EWOULDBLOCK, _, _) when Sys.os_type = "Win32" ->
+           [{ priority = priority ; events = [`Write unix_socket] ;
+              handler = f pos }]
        | Timeout_exc -> on_error Timeout; []
        | Unix.Unix_error(x,y,z) -> on_error (Unix(x,y,z)); []
        | e -> on_error (Unknown e); []
-    in  
+    in
+    let task = {
+      priority = priority ;
+      events = events ;
+      handler  = (f 0)
+    } in
     if length > 0 then
-        let task = 
-          {
-            priority = priority ;
-            events = events ;
-            handler  = (f 0)
-          }
-        in
+      (* Win32 is particularly bad with writting on sockets. It is nearly impossible
+       * to write proper non-blocking code. send will block on blocking sockets if
+       * there isn't enough data available instead of returning a partial buffer
+       * and WSAEventSelect will not return if the socket still has available space.
+       * Thus, setting the socket to non-blocking and writting as much as we can. *)
+      if Sys.os_type = "Win32" then
+       begin
+        Unix.set_nonblock unix_socket ;
+        List.iter (add scheduler) (f 0 [`Write unix_socket])
+       end
+      else
         add scheduler task
     else 
         exec ()
